#ifndef dosutil_h
#define dosutil_h

#include <setjmp.h>
#include <sys/types.h>

typedef struct oldenvbuf
{
  int fds[3];

  jmp_buf top_level;
  jmp_buf subshell_top_level;

  char *pwd;
  char *oldpwd;
  char *curdir;

  int interactive;
  int interactive_shell;
  int login_shell;
  int subshell_environment;
  int subshell_exit_builtin;
  int exit_immediately_on_error;
} OLDENVBUF;

extern int dosutil_path_separator;
extern int dosutil_path_slash;
extern char dosutil_sysroot[];
extern int dosutil_sysroot_len;
extern int dosutil_wait_status;
extern pid_t dosutil_current_pid;

char *dosutil_make_import_path (char *value);
char *dosutil_make_export_path (char *value);
void dosutil_import_envvar (void);
char *dosutil_addslash (char *path);
char *dosutil_delslash (char *path);
char *dosutil_toslash (char *path);
char *dosutil_tobslash (char *path);
char *dosutil_tolower (char *path);
char *dosutil_toupper (char *path);
char *dosutil_delsysroot (char *path);
char *dosutil_expand_path (char *path);
char *dosutil_fix_variable (char *path);
char **dosutil_get_import_variables (void);
void dosutil_put_export_variables (char **vp);
void dosutil_reserve_fds (void);
void dosutil_save_cwd (void);
void dosutil_save_std_fds (int fds[]);
void dosutil_restore_std_fds (int fds[]);
int dosutil_save_all_environment (OLDENVBUF *envp);
int dosutil_restore_all_environment (OLDENVBUF *envp);

void dosutil_pipe_clean (int no);
int dosutil_pipe_no (void);
int dosutil_pipe (int fds[]);
pid_t dosutil_make_child_pid (void);

int dosutil_chdir (char *path);
char *dosutil_getcwd (char *buffer, int length);
char *dosutil_getwd (char *buffer);

/* define in trap.c */
int dosutil_get_signal_modes (int sig);
int dosutil_set_signal_modes (int sig, int modes);

#if defined (__DJGPP__)
#include <dpmi.h>
#define DOSUTIL_EXCEPTION_COUNT 18

typedef struct oldexceptbuf
{
  __dpmi_paddr except_original[DOSUTIL_EXCEPTION_COUNT];
  __dpmi_paddr kbd_original;
  __dpmi_paddr npx_original;
  __dpmi_raddr cbrk_original;
} OLDEXCEPTBUF;
char *dosutil_add_executable_suffix (char *name, char *ext);
void dosutil_save_exceptions (OLDEXCEPTBUF *exceptionp);
void dosutil_restore_exceptions (OLDEXCEPTBUF *exceptionp);
int *dosutil_check_dpmi_selectors (void);
void dosutil_free_dpmi_selectors (int *selectors);
#endif /* !__DJGPP__ */

#endif /* dosutil_h */
